/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2001
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dlfcn.h>

#include <cups/cups.h>
#include <cups/ppd.h>
#include <cups/raster.h>

#include "Device.hpp"

int
main (int argc, char *argv[])
{
#define MAX_SCAN_LINES 1024
   const char         *pszPPDFilename;        /* PPD environment variable */
   ppd_file_t         *ppd;                   /* PPD file */
   int                 fdInput;               /* File descriptor */
   cups_raster_t      *pRasterStream;         /* Raster stream to read from */
   cups_page_header_t  pageHeader;            /* Page header from file */
   void               *vhDevice          = 0;
   Device             *pDevice           = 0;
   PFNNEWDEVICEWARGS   pfnNewDeviceWArgs = 0;
   char               *pszJobProperties  = 0;
   int                 cbJobProperties   = 0;
   int                 iPageNumber       = 0;
   unsigned int        uiScanline        = 0;
   int                 iNumScanLines     = MAX_SCAN_LINES;
   PBYTE               pbBits            = 0;
   PBITMAPINFO2        pbmi2Bitmap       = 0;
   int                 iNumColors        = 0;
   int                 icbBitmapScanLine = 0;
   int                 iBytesToAllocate  = 0;
   RECTL               rectlPageLocation;

   if (  argc < 6
      || argc > 7
      )
   {
      /*
       * We don't have the correct number of arguments; write an error message
       * and return.
       */

      fputs ("ERROR: CUPSToOmni job-id user title copies options [file]\n", stderr);
      return 1;
   }

   /*
    * Get the PPD file and figure out which driver to use...
    */

   if ((pszPPDFilename = getenv ("PPD")) == NULL)
   {
      fputs ("ERROR: Fatal error: PPD environment variable not set!\n", stderr);
      return 1;
   }

   if ((ppd = ppdOpenFile (pszPPDFilename)) == NULL)
   {
      fprintf (stderr,
               "ERROR: Fatal error: Unable to load PPD file \"%s\"!\n",
               pszPPDFilename);

      return 1;
   }

   if (ppd->nickname == NULL)
   {
      fprintf (stderr,
               "ERROR: Fatal error: No NickName attribute in PPD file \"%s\"!\n",
               pszPPDFilename);
      ppdClose (ppd);

      return 1;
   }

   vhDevice = dlopen (ppd->nickname, RTLD_NOW | RTLD_GLOBAL);

   if (!vhDevice)
   {
      cerr << "Error: unable to load " << ppd->nickname << endl;
      cerr << "dlerror returns " << dlerror () << endl;

      ppdClose (ppd);

      return 1;
   }

   pfnNewDeviceWArgs = (PFNNEWDEVICEWARGS)dlsym (vhDevice, "newDevice__FPcb");

   if (!pfnNewDeviceWArgs)
   {
      cerr << "Error: " << ppd->nickname << " is not a valid library" << endl;

      dlclose (vhDevice);
      ppdClose (ppd);

      return 1;
   }

   ppdClose (ppd);
   ppd = 0;

   /*
    * Open the page stream...
    */
   if (argc == 7)
   {
      if ((fdInput = open (argv[6], O_RDONLY)) == -1)
      {
         perror ("ERROR: Unable to open raster file - ");
         sleep (1);

         return 1;
      }
   }
   else
   {
      fdInput = 0;
   }

   pRasterStream = cupsRasterOpen (fdInput, CUPS_RASTER_READ);

   iPageNumber = 0;

   /*
    * Process pages as needed...
    */
   while (cupsRasterReadHeader (pRasterStream, &pageHeader))
   {
      cbJobProperties = 0;

      // @TBD
///// "form = " ????????????????????????
///// "tray = " pageHeader.MediaClass
///// "media = " pageHeader.MediaType
///// "resolution = " pageHeader.cupsCompression
///// "dither = " pageHeader.OutputType

      pDevice = pfnNewDeviceWArgs (pszJobProperties, false);

      if (!pDevice)
      {
         cerr << "Error: Could not instantiate a new device." << endl;

         dlclose (vhDevice);
         return 1;
      }

      iNumScanLines = min (MAX_SCAN_LINES, (int)(pageHeader.cupsHeight - uiScanline));
      iNumColors    = 1 << pageHeader.cupsBitsPerPixel;

      if (!pbmi2Bitmap)
      {
         // Allocate a full bitmap info 2 and bitmap header message
         iBytesToAllocate = sizeof (BITMAPINFO2);
         if (256 >= iNumColors)
            iBytesToAllocate += (iNumColors - 1) * sizeof (RGB2);

         pbmi2Bitmap = (PBITMAPINFO2)calloc (1, iBytesToAllocate);
         if (!pbmi2Bitmap)
            goto done;

         // Initialize it
         pbmi2Bitmap->cbFix         = sizeof (BITMAPINFO2)
                                    - sizeof (((PBITMAPINFO2)0)->argbColor)
                                     ;
         pbmi2Bitmap->cx            = pageHeader.cupsWidth;
         pbmi2Bitmap->cy            = iNumScanLines;
         pbmi2Bitmap->cPlanes       = 1;
         pbmi2Bitmap->cBitCount     = pageHeader.cupsBitsPerPixel;
         pbmi2Bitmap->cclrUsed      = 1 << pbmi2Bitmap->cBitCount;
         pbmi2Bitmap->cclrImportant = 0;                            // All are important

         // Read the color table, if there is one
         // @TBD
      }

      icbBitmapScanLine = ((pbmi2Bitmap->cx * pbmi2Bitmap->cBitCount + 31) >> 5) << 2;

      if (!pbBits)
      {
         iBytesToAllocate = icbBitmapScanLine * iNumScanLines;

         pbBits = (PBYTE)calloc (1, iBytesToAllocate);
         if (!pbBits)
            goto done;
      }

      iPageNumber++;
      uiScanline = 0;

      if (1 == iPageNumber)
      {
         pDevice->beginJob ();
      }
      else
      {
         pDevice->newFrame ();
      }

      // @TBD - loop getting at most MAX_SCAN_LINES at a time
      while (1)
      {
      }

      while (uiScanline < pageHeader.cupsHeight)
      {
         iNumScanLines = min (MAX_SCAN_LINES, (int)(pageHeader.cupsHeight - uiScanline));
         pbmi2Bitmap->cy = iNumScanLines;

         rectlPageLocation.xLeft   = 0;
         rectlPageLocation.yBottom = uiScanline;
         rectlPageLocation.xRight  = pbmi2Bitmap->cx;
         rectlPageLocation.yTop    = uiScanline + iNumScanLines - 1;

         for (int y = 0; y < iNumScanLines; y++)
         {
            cupsRasterReadPixels (pRasterStream,
                                  pbBits + y * icbBitmapScanLine,
                                  pageHeader.cupsBytesPerLine);
            uiScanline++;
         }

         // Send it to the device
         pDevice->rasterize (pbBits,
                             pbmi2Bitmap,
                             &rectlPageLocation,
                             BITBLT_BITMAP);
      }
   }

   pDevice->endJob ();

done:
   /*
    * Close the raster stream...
    */
   cupsRasterClose (pRasterStream);

   if (fdInput != 0)
   {
      close (fdInput);
   }

   if (pbmi2Bitmap)
   {
      free (pbmi2Bitmap);
      pbmi2Bitmap = 0;
   }

   delete pDevice;

   if (vhDevice)
   {
      dlclose (vhDevice);
      vhDevice = 0;
   }

   if (0 == iPageNumber)
   {
      fputs ("ERROR: No pages found!\n", stderr);
   }
   else
   {
      fputs ("INFO: Ready to print.\n", stderr);
   }
}
